@@ -17,6 +17,7 @@ from kuryr.lib import constants as kl_const
 from kuryr.lib import exceptions as kl_exc
 from kuryr.lib import segmentation_type_drivers as seg_driver
 from openstack import exceptions as os_exc
+from oslo_config import cfg
 from oslo_log import log as logging
 
 from kuryr_kubernetes import clients
@@ -32,6 +33,9 @@ LOG = logging.getLogger(__name__)
 
 DEFAULT_MAX_RETRY_COUNT = 3
 DEFAULT_RETRY_INTERVAL = 1
+ACTIVE_TIMEOUT = 90
+
+CONF = cfg.CONF
 
 
 class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
@@ -45,13 +49,14 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
         rq = self._get_port_request(pod, project_id, subnets, security_groups)
         port = os_net.create_port(**rq)
         self._check_port_binding([port])
-        utils.tag_neutron_resources([port])
+        if not self._tag_on_creation:
+            utils.tag_neutron_resources([port])
         vlan_id = self._add_subport(trunk_id, port.id)
 
         return ovu.neutron_to_osvif_vif_nested_vlan(port, subnets, vlan_id)
 
     def request_vifs(self, pod, project_id, subnets, security_groups,
-                     num_ports, trunk_ip=None):
+                     num_ports, semaphore, trunk_ip=None):
         """This method creates subports and returns a list with their vifs.
 
         It creates up to num_ports subports and attaches them to the trunk
@@ -80,14 +85,20 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
             LOG.error("There are no vlan ids available to create subports")
             return []
 
-        bulk_port_rq = {'ports': [port_rq] * len(subports_info)}
-        try:
-            ports = list(os_net.create_ports(bulk_port_rq))
-        except os_exc.SDKException:
-            LOG.exception("Error creating bulk ports: %s", bulk_port_rq)
-            raise
+        bulk_port_rq = [port_rq] * len(subports_info)
+        # restrict amount of create Ports in bulk that might be running
+        # in parallel.
+        with semaphore:
+            try:
+                ports = list(os_net.create_ports(bulk_port_rq))
+            except os_exc.SDKException:
+                for subport_info in subports_info:
+                    self._release_vlan_id(subport_info['segmentation_id'])
+                LOG.exception("Error creating bulk ports: %s", bulk_port_rq)
+                raise
         self._check_port_binding(ports)
-        utils.tag_neutron_resources(ports)
+        if not self._tag_on_creation:
+            utils.tag_neutron_resources(ports)
 
         for index, port in enumerate(ports):
             subports_info[index]['port_id'] = port['id']
@@ -97,13 +108,15 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
                 os_net.add_trunk_subports(trunk_id, subports_info)
             except os_exc.ConflictException:
                 LOG.error("vlan ids already in use on trunk")
-                for port in ports:
-                    os_net.delete_port(port.id)
+                utils.delete_ports(ports)
+                for subport_info in subports_info:
+                    self._release_vlan_id(subport_info['segmentation_id'])
                 return []
         except os_exc.SDKException:
             LOG.exception("Error happened during subport addition to trunk")
-            for port in ports:
-                os_net.delete_port(port.id)
+            utils.delete_ports(ports)
+            for subport_info in subports_info:
+                self._release_vlan_id(subport_info['segmentation_id'])
             return []
 
         vifs = []
@@ -113,11 +126,45 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
             vifs.append(vif)
         return vifs
 
+    def activate_vif(self, vif, pod=None, retry_info=None):
+        try:
+            super().activate_vif(vif)
+        except k_exc.PortNotReady:
+            if retry_info and retry_info.get('elapsed', 0) > ACTIVE_TIMEOUT:
+                parent_port = self._get_parent_port(pod)
+                trunk_id = self._get_trunk_id(parent_port)
+                # NOTE(dulek): We don't need a lock to prevent VLAN ID from
+                #              being taken over because the SegmentationDriver
+                #              will keep it reserved in memory unless we
+                #              release it. And we won't.
+                LOG.warning('Subport %s is in DOWN status for more than %d '
+                            'seconds. This is a Neutron issue. Attempting to '
+                            'reattach the subport to trunk %s using VLAN ID %s'
+                            ' to fix it.', vif.id, retry_info['elapsed'],
+                            trunk_id, vif.vlan_id)
+                try:
+                    self._remove_subport(trunk_id, vif.id)
+                except os_exc.NotFoundException:
+                    # NOTE(dulek): This may happen when _add_subport() failed
+                    #              or Kuryr crashed between the calls. Let's
+                    #              try to fix it hoping that VLAN ID is still
+                    #              free.
+                    LOG.warning('Subport %s was not attached to the trunk. '
+                                'Trying to attach it anyway.', vif.id)
+                self._add_subport(trunk_id, vif.id,
+                                  requested_vlan_id=vif.vlan_id)
+                LOG.warning("Reattached subport %s, its state will be "
+                            "rechecked when event will be retried.", vif.id)
+            raise
+
     def release_vif(self, pod, vif, project_id=None, security_groups=None):
         os_net = clients.get_network_client()
         parent_port = self._get_parent_port(pod)
         trunk_id = self._get_trunk_id(parent_port)
-        self._remove_subport(trunk_id, vif.id)
+        try:
+            self._remove_subport(trunk_id, vif.id)
+        except os_exc.NotFoundException:
+            pass
         self._release_vlan_id(vif.vlan_id)
         os_net.delete_port(vif.id)
 
@@ -139,6 +186,11 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
         if security_groups:
             port_req_body['security_groups'] = security_groups
 
+        if self._tag_on_creation:
+            tags = CONF.neutron_defaults.resource_tags
+            if tags:
+                port_req_body['tags'] = tags
+
         return port_req_body
 
     def _create_subports_info(self, pod, project_id, subnets,
@@ -149,7 +201,7 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
         in_use_vlan_ids = self._get_in_use_vlan_ids_set(trunk_id)
         port_rq = self._get_port_request(pod, project_id, subnets,
                                          security_groups, unbound)
-        for i in range(num_ports):
+        for _ in range(num_ports):
             try:
                 vlan_id = seg_driver.allocate_segmentation_id(in_use_vlan_ids)
             except kl_exc.SegmentationIdAllocationFailure:
@@ -172,7 +224,7 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
                       "with a Neutron vlan trunk")
             raise k_exc.K8sNodeTrunkPortFailure
 
-    def _add_subport(self, trunk_id, subport):
+    def _add_subport(self, trunk_id, subport, requested_vlan_id=None):
         """Adds subport port to Neutron trunk
 
         This method gets vlanid allocated from kuryr segmentation driver.
@@ -186,30 +238,36 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
         os_net = clients.get_network_client()
         retry_count = 1
         while True:
-            try:
-                vlan_id = self._get_vlan_id(trunk_id)
-            except os_exc.SDKException:
-                LOG.error("Getting VlanID for subport on "
-                          "trunk %s failed!!", trunk_id)
-                raise
+            if requested_vlan_id:
+                vlan_id = requested_vlan_id
+            else:
+                try:
+                    vlan_id = self._get_vlan_id(trunk_id)
+                except os_exc.SDKException:
+                    LOG.error("Getting VLAN ID for subport on "
+                              "trunk %s failed!!", trunk_id)
+                    raise
+
             subport = [{'segmentation_id': vlan_id,
                         'port_id': subport,
-                       'segmentation_type': 'vlan'}]
+                        'segmentation_type': 'vlan'}]
             try:
                 os_net.add_trunk_subports(trunk_id, subport)
             except os_exc.ConflictException:
-                if retry_count < DEFAULT_MAX_RETRY_COUNT:
-                    LOG.error("vlanid already in use on trunk, "
-                              "%s. Retrying...", trunk_id)
+                if (retry_count < DEFAULT_MAX_RETRY_COUNT and
+                        not requested_vlan_id):
+                    LOG.error("VLAN ID already in use on trunk %s. "
+                              "Retrying.", trunk_id)
                     retry_count += 1
                     sleep(DEFAULT_RETRY_INTERVAL)
                     continue
                 else:
-                    LOG.error(
-                        "MAX retry count reached. Failed to add subport")
+                    LOG.error("Failed to add subport %s to trunk %s due to "
+                              "VLAN ID %d conflict.", subport, trunk_id,
+                              vlan_id)
                     raise
-
             except os_exc.SDKException:
+                self._release_vlan_id(vlan_id)
                 LOG.exception("Error happened during subport "
                               "addition to trunk %s", trunk_id)
                 raise
@@ -222,6 +280,12 @@ class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
             subports_body.append({'port_id': subport_id})
         try:
             os_net.delete_trunk_subports(trunk_id, subports_body)
+        except os_exc.NotFoundException:
+            if len(subports_id) > 1:
+                LOG.debug('Not Found on subport deletion, most likely some '
+                          'subports in the list got detached already.')
+                raise  # We don't know if all ports are detached, so raise.
+            # If single requested port is detached already, we're cool, ignore.
         except os_exc.SDKException:
             LOG.exception("Error happened during subport removal from "
                           "trunk %s", trunk_id)
