@@ -15,10 +15,10 @@
 
 import abc
 import collections
-import eventlet
 import os
-import time
+import threading
 
+import eventlet
 from kuryr.lib._i18n import _
 from kuryr.lib import constants as kl_const
 from openstack import exceptions as os_exc
@@ -78,8 +78,10 @@ vif_pool_driver_opts = [
 oslo_cfg.CONF.register_opts(vif_pool_driver_opts, "vif_pool")
 
 node_vif_driver_caching_opts = [
-    oslo_cfg.BoolOpt('caching', default=True),
-    oslo_cfg.IntOpt('cache_time', default=3600),
+    oslo_cfg.BoolOpt('caching', default=True,
+                     help=_('Enable caching of vifs.')),
+    oslo_cfg.IntOpt('cache_time', default=3600,
+                    help=_('TTL, in seconds, for cached vifs')),
 ]
 
 oslo_cfg.CONF.register_opts(node_vif_driver_caching_opts,
@@ -97,12 +99,13 @@ VIF_TYPE_TO_DRIVER_MAPPING = {
     'VIFBridge': 'neutron-vif',
     'VIFVlanNested': 'nested-vlan',
     'VIFMacvlanNested': 'nested-macvlan',
-    'VIFSriov': 'sriov',
     'VIFDPDKNested': 'nested-dpdk',
     'VIFVHostUser': 'neutron-vif',
 }
 
 NODE_PORTS_CLEAN_FREQUENCY = 600  # seconds
+POPULATE_POOL_TIMEOUT = 420  # seconds
+BULK_PORTS_CREATION_REQUESTS = 20
 
 
 class NoopVIFPool(base.VIFPoolDriver):
@@ -118,8 +121,8 @@ class NoopVIFPool(base.VIFPoolDriver):
     def release_vif(self, pod, vif, *argv):
         self._drv_vif.release_vif(pod, vif, *argv)
 
-    def activate_vif(self, vif):
-        self._drv_vif.activate_vif(vif)
+    def activate_vif(self, vif, **kwargs):
+        self._drv_vif.activate_vif(vif, **kwargs)
 
     def update_vif_sgs(self, pod, sgs):
         self._drv_vif.update_vif_sgs(pod, sgs)
@@ -141,9 +144,8 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
     are the 'port_id' and the values are the vif objects.
     _recyclable_ports is a dictionary with the Neutron ports to be
     recycled. The keys are the 'port_id' and their values are the 'pool_key'.
-    _last_update is a dictionary with the timestamp of the last population
-    action for each pool. The keys are the pool_keys and the values are the
-    timestamps.
+    _populate_pool_lock is a dict with the pool_key as key and a lock as value.
+    Also, there is a _lock to control access to _populate_pool_lock dict.
 
     The following driver configuration options exist:
     - ports_pool_max: it specifies how many ports can be kept at each pool.
@@ -155,7 +157,9 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
     - ports_pool_batch: target number of ports to be created in bulk requests
     when populating pools.
     - ports_pool_update_frequency: interval in seconds between ports pool
-    updates, both for populating pools as well as for recycling ports.
+    updates for recycling ports.
+    Also, it has a Semaphore _create_ports_semaphore to restrict the number of
+    bulk Ports creation calls running in parallel.
     """
 
     def __init__(self):
@@ -168,8 +172,8 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
     def set_vif_driver(self, driver):
         self._drv_vif = driver
 
-    def activate_vif(self, vif):
-        self._drv_vif.activate_vif(vif)
+    def activate_vif(self, vif, **kwargs):
+        self._drv_vif.activate_vif(vif, **kwargs)
 
     def update_vif_sgs(self, pod, sgs):
         self._drv_vif.update_vif_sgs(pod, sgs)
@@ -205,65 +209,111 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
 
         pool_key = self._get_pool_key(host_addr, project_id, None, subnets)
 
+        # NOTE(maysams): It's possible that more recent Pods will retrieve
+        # the Ports from the pool that older Pods were waiting for. In case
+        # this happens, the event will be retried.
         try:
             return self._get_port_from_pool(pool_key, pod, subnets,
                                             tuple(sorted(security_groups)))
         except exceptions.ResourceNotReady:
             LOG.debug("Ports pool does not have available ports: %s", pool_key)
-            # NOTE(dulek): We're passing raise_not_ready=False because this
-            #              will be run outside of handlers thread, so raising
-            #              it will only result in an ugly log from eventlet.
-            eventlet.spawn(self._populate_pool, pool_key, pod, subnets,
-                           tuple(sorted(security_groups)),
-                           raise_not_ready=False)
+            if self._populate_pool(pool_key, pod, subnets,
+                                   tuple(sorted(security_groups))):
+                return self._get_port_from_pool(
+                    pool_key, pod, subnets, tuple(sorted(security_groups)))
             raise
 
-    def _get_port_from_pool(self, pool_key, pod, subnets, security_groups):
+    def _set_port_debug(self, port_id, pod):
+        """_set_port_debug sets name to the port to simplify debugging"""
         raise NotImplementedError()
 
-    def _populate_pool(self, pool_key, pod, subnets, security_groups,
-                       raise_not_ready=True):
+    def _get_populate_pool_lock(self, pool_key):
+        with self._lock:
+            return self._populate_pool_lock[pool_key]
+
+    def _get_port_from_pool(self, pool_key, pod, subnets, security_groups):
+        try:
+            pool_ports = self._available_ports_pools[pool_key]
+        except (KeyError, AttributeError):
+            raise exceptions.ResourceNotReady(pod)
+
+        try:
+            port_id = pool_ports[security_groups].pop()
+        except (KeyError, IndexError):
+            # Get another port from the pool and update the SG to the
+            # appropriate one. It uses a port from the group that was updated
+            # longer ago - these will be at the front of the OrderedDict.
+            for sg_group, ports in pool_ports.items():
+                try:
+                    port_id = pool_ports[sg_group].pop()
+                    break
+                except (IndexError, KeyError):
+                    continue
+            else:
+                # pool is empty, no port to reuse
+                raise exceptions.ResourceNotReady(pod)
+            os_net = clients.get_network_client()
+            os_net.update_port(port_id, security_groups=list(security_groups))
+        if config.CONF.kubernetes.port_debug:
+            self._set_port_debug(port_id, pod)
+        eventlet.spawn(self._populate_pool, pool_key, pod, subnets,
+                       security_groups)
+        # Add protection from port_id not in existing_vifs
+        try:
+            port = self._existing_vifs[port_id]
+        except KeyError:
+            LOG.debug('Missing port on existing_vifs, this should not happen.'
+                      ' Retrying.')
+            raise exceptions.ResourceNotReady(pod)
+        return port
+
+    def _populate_pool(self, pool_key, pod, subnets, security_groups):
         # REVISIT(ltomasbo): Drop the subnets parameter and get the information
         # from the pool_key, which will be required when multi-network is
         # supported
+        kubernetes = clients.get_kubernetes_client()
+
         if not self._recovered_pools:
             LOG.debug("Kuryr-controller not yet ready to populate pools.")
-            if raise_not_ready:
-                raise exceptions.ResourceNotReady(pod)
-            else:
-                return
-        now = time.time()
-        last_update = 0
-        pool_updates = self._last_update.get(pool_key)
-        if pool_updates:
-            last_update = pool_updates.get(security_groups, 0)
+            return False
+        ports_pool_min = oslo_cfg.CONF.vif_pool.ports_pool_min
+        lock = self._get_populate_pool_lock(pool_key)
+        # NOTE(maysams): Only allow one request vifs per pool and times out
+        # if takes 420 sec.
+        if lock.acquire(timeout=POPULATE_POOL_TIMEOUT):
+            pool_size = self._get_pool_size(pool_key)
             try:
-                if (now - oslo_cfg.CONF.vif_pool.ports_pool_update_frequency <
-                        last_update):
-                    LOG.debug("Not enough time since the last pool update")
-                    return
-            except AttributeError:
-                LOG.debug("Kuryr-controller not yet ready to populate pools.")
-                return
-        self._last_update[pool_key] = {security_groups: now}
+                if pool_size < ports_pool_min:
+                    num_ports = max(oslo_cfg.CONF.vif_pool.ports_pool_batch,
+                                    ports_pool_min - pool_size)
+                    try:
+                        vifs = self._drv_vif.request_vifs(
+                            pod=pod,
+                            project_id=pool_key[1],
+                            subnets=subnets,
+                            security_groups=security_groups,
+                            num_ports=num_ports,
+                            semaphore=self._create_ports_semaphore)
+                    except os_exc.SDKException as exc:
+                        kubernetes.add_event(
+                            pod, 'FailToPopulateVIFPool',
+                            f'There was an error during populating VIF pool '
+                            f'for pod: {exc.message}', type_='Warning')
+                        raise
 
-        pool_size = self._get_pool_size(pool_key)
-        if pool_size < oslo_cfg.CONF.vif_pool.ports_pool_min:
-            num_ports = max(oslo_cfg.CONF.vif_pool.ports_pool_batch,
-                            oslo_cfg.CONF.vif_pool.ports_pool_min - pool_size)
-            vifs = self._drv_vif.request_vifs(
-                pod=pod,
-                project_id=pool_key[1],
-                subnets=subnets,
-                security_groups=security_groups,
-                num_ports=num_ports)
-            for vif in vifs:
-                self._existing_vifs[vif.id] = vif
-                self._available_ports_pools.setdefault(
-                    pool_key, {}).setdefault(
-                        security_groups, []).append(vif.id)
-            if not vifs:
-                self._last_update[pool_key] = {security_groups: last_update}
+                    for vif in vifs:
+                        self._existing_vifs[vif.id] = vif
+                        self._available_ports_pools[pool_key].setdefault(
+                                security_groups, []).append(vif.id)
+                    if vifs:
+                        # Mark it as updated most recently.
+                        self._available_ports_pools[pool_key].move_to_end(
+                            security_groups)
+            finally:
+                lock.release()
+        else:
+            return False
+        return True
 
     def release_vif(self, pod, vif, project_id, security_groups,
                     host_addr=None):
@@ -294,9 +344,9 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
         kubernetes = clients.get_kubernetes_client()
         in_use_ports = []
         networks = {}
-        running_pods = kubernetes.get(constants.K8S_API_BASE + '/pods')
-        for pod in running_pods['items']:
-            vifs = c_utils.get_vifs(pod)
+        kuryr_ports = kubernetes.get(constants.K8S_API_CRD_KURYRPORTS)
+        for kp in kuryr_ports['items']:
+            vifs = c_utils.get_vifs(kp)
             for data in vifs.values():
                 in_use_ports.append(data.id)
                 networks[data.network.id] = data.network
@@ -332,11 +382,12 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
                     os_net.update_port(port_id, security_groups=None)
                     # add the port to the default pool
                     self._available_ports_pools[pool_key].setdefault(
-                        tuple([]), []).append(port_id)
+                        (), []).append(port_id)
                 # NOTE(ltomasbo): as this ports were not created for this
                 # pool, ensuring they are used first, marking them as the
                 # most outdated
-                self._last_update[pool_key] = {tuple([]): 0}
+                self._available_ports_pools[pool_key].move_to_end(
+                    (), last=False)
 
     def _create_healthcheck_file(self):
         # Note(ltomasbo): Create a health check file when the pre-created
@@ -360,10 +411,14 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
         except OSError:
             pass
 
-        self._available_ports_pools = collections.defaultdict()
+        self._available_ports_pools = collections.defaultdict(
+            collections.OrderedDict)
         self._existing_vifs = collections.defaultdict()
         self._recyclable_ports = collections.defaultdict()
-        self._last_update = collections.defaultdict()
+        self._lock = threading.Lock()
+        self._populate_pool_lock = collections.defaultdict(threading.Lock)
+        semaphore = eventlet.semaphore.Semaphore(BULK_PORTS_CREATION_REQUESTS)
+        self._create_ports_semaphore = semaphore
 
     def _get_trunks_info(self):
         """Returns information about trunks and their subports.
@@ -432,10 +487,9 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
                         # getting the Network and Subnet info from
                         # Network defined on an existing KuryrPort CR.
                         # This assumes only one Subnet exists per Network.
-                        if in_use_networks.get(port.network_id):
-                            subnets[subnet_id] = {
-                                subnet_id: in_use_networks.get(
-                                    port.network_id)}
+                        network = in_use_networks.get(port.network_id)
+                        if network:
+                            subnets[subnet_id] = {subnet_id: network}
                         else:
                             subnets[subnet_id] = {
                                 subnet_id: utils.get_subnet(subnet_id)}
@@ -473,9 +527,8 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
                                       "Skipping.", port.id)
                             continue
         else:
-            for port in existing_ports:
-                if not port.binding_host_id:
-                    os_net.delete_port(port.id)
+            c_utils.delete_ports([p for p in existing_ports
+                                  if not p.binding_host_id])
 
     def _cleanup_removed_nodes(self):
         """Remove ports associated to removed nodes."""
@@ -549,12 +602,8 @@ class BaseVIFPool(base.VIFPoolDriver, metaclass=abc.ABCMeta):
                     del self._existing_vifs[subport.id]
                 except KeyError:
                     LOG.debug('Port %s is not in the ports list.', subport.id)
-                try:
-                    os_net.delete_port(subport.id)
-                except os_exc.SDKException:
-                    LOG.debug("Problem deleting leftover port %s. "
-                              "Skipping.", subport.id)
-                else:
+                port_deleted = c_utils.delete_port(subport)
+                if port_deleted:
                     previous_ports_to_remove.remove(subport.id)
 
             # normal ports, or subports not yet attached
@@ -595,56 +644,10 @@ class NeutronVIFPool(BaseVIFPool):
     def _get_host_addr(self, pod):
         return pod['spec']['nodeName']
 
-    def _get_port_from_pool(self, pool_key, pod, subnets, security_groups):
-        try:
-            pool_ports = self._available_ports_pools[pool_key]
-        except (KeyError, AttributeError):
-            raise exceptions.ResourceNotReady(pod)
-        try:
-            port_id = pool_ports[security_groups].pop()
-        except (KeyError, IndexError):
-            # Get another port from the pool and update the SG to the
-            # appropriate one. It uses a port from the group that was updated
-            # longer ago
-            pool_updates = self._last_update.get(pool_key, {})
-            if not pool_updates:
-                # No pools update info. Selecting a random one
-                for sg_group, ports in list(pool_ports.items()):
-                    if len(ports) > 0:
-                        port_id = pool_ports[sg_group].pop()
-                        break
-                else:
-                    raise exceptions.ResourceNotReady(pod)
-            else:
-                min_date = -1
-                for sg_group, date in list(pool_updates.items()):
-                    if pool_ports.get(sg_group):
-                        if min_date == -1 or date < min_date:
-                            min_date = date
-                            min_sg_group = sg_group
-                if min_date == -1:
-                    # pool is empty, no port to reuse
-                    raise exceptions.ResourceNotReady(pod)
-                port_id = pool_ports[min_sg_group].pop()
-            os_net = clients.get_network_client()
-            os_net.update_port(port_id, security_groups=list(security_groups))
-        if config.CONF.kubernetes.port_debug:
-            os_net = clients.get_network_client()
-            os_net.update_port(port_id, name=c_utils.get_port_name(pod),
-                               device_id=pod['metadata']['uid'])
-        # check if the pool needs to be populated
-        if (self._get_pool_size(pool_key) <
-                oslo_cfg.CONF.vif_pool.ports_pool_min):
-            eventlet.spawn(self._populate_pool, pool_key, pod, subnets,
-                           security_groups)
-        # Add protection from port_id not in existing_vifs
-        try:
-            port = self._existing_vifs[port_id]
-        except KeyError:
-            LOG.debug('Missing port on existing_vifs, this should not happen.'
-                      ' Retrying.')
-            raise exceptions.ResourceNotReady(pod)
-        return port
+    def _set_port_debug(self, port_id, pod):
+        os_net = clients.get_network_client()
+        os_net.update_port(port_id, name=c_utils.get_port_name(pod),
+                           device_id=pod['metadata']['uid'])
 
     def _return_ports_to_pool(self):
         """Recycle ports to be reused by future pods.
@@ -702,9 +705,11 @@ class NeutronVIFPool(BaseVIFPool):
                                     "reused, put back on the cleanable "
                                     "pool.", port_id)
                         continue
-                self._available_ports_pools.setdefault(
-                    pool_key, {}).setdefault(
-                        sg_current.get(port_id), []).append(port_id)
+                sg = sg_current.get(port_id)
+                self._available_ports_pools[pool_key].setdefault(
+                        sg, []).append(port_id)
+                # Move it to the end of ports to update the SG.
+                self._available_ports_pools[pool_key].move_to_end(sg)
             else:
                 try:
                     del self._existing_vifs[port_id]
@@ -763,8 +768,7 @@ class NeutronVIFPool(BaseVIFPool):
                                           net_obj.id, None)
 
             self._existing_vifs[port.id] = vif
-            self._available_ports_pools.setdefault(
-                pool_key, {}).setdefault(
+            self._available_ports_pools[pool_key].setdefault(
                     tuple(sorted(port.security_group_ids)), []).append(port.id)
 
         LOG.info("PORTS POOL: pools updated with pre-created ports")
@@ -775,7 +779,8 @@ class NeutronVIFPool(BaseVIFPool):
             LOG.debug("Kuryr-controller not yet ready to delete network "
                       "pools.")
             raise exceptions.ResourceNotReady(net_id)
-        os_net = clients.get_network_client()
+
+        epool = eventlet.GreenPool(constants.LEFTOVER_RM_POOL_SIZE)
 
         # NOTE(ltomasbo): Note the pods should already be deleted, but their
         # associated ports may not have been recycled yet, therefore not being
@@ -795,9 +800,20 @@ class NeutronVIFPool(BaseVIFPool):
                     LOG.debug('Port %s is not in the ports list.', port_id)
                 # NOTE(gryf): openstack client doesn't return information, if
                 # the port deos not exists
-                os_net.delete_port(port_id)
 
-            self._available_ports_pools[pool_key] = {}
+            # Delete ports concurrently
+            for result in epool.imap(c_utils.delete_neutron_port, ports_id):
+                if result:
+                    LOG.error('During Neutron port deletion an error occured: '
+                              '%s', result)
+                    raise result
+
+            del self._available_ports_pools[pool_key]
+            with self._lock:
+                try:
+                    del self._populate_pool_lock[pool_key]
+                except KeyError:
+                    pass
 
 
 class NestedVIFPool(BaseVIFPool):
@@ -861,56 +877,9 @@ class NestedVIFPool(BaseVIFPool):
         super(NestedVIFPool, self).release_vif(
             pod, vif, project_id, security_groups, host_addr=host_addr)
 
-    def _get_port_from_pool(self, pool_key, pod, subnets, security_groups):
-        try:
-            pool_ports = self._available_ports_pools[pool_key]
-        except (KeyError, AttributeError):
-            raise exceptions.ResourceNotReady(pod)
-
+    def _set_port_debug(self, port_id, pod):
         os_net = clients.get_network_client()
-
-        try:
-            port_id = pool_ports[security_groups].pop()
-        except (KeyError, IndexError):
-            # Get another port from the pool and update the SG to the
-            # appropriate one. It uses a port from the group that was updated
-            # longer ago
-            pool_updates = self._last_update.get(pool_key, {})
-            if not pool_updates:
-                # No pools update info. Selecting a random one
-                for sg_group, ports in list(pool_ports.items()):
-                    if len(ports) > 0:
-                        port_id = pool_ports[sg_group].pop()
-                        break
-                else:
-                    raise exceptions.ResourceNotReady(pod)
-            else:
-                min_date = -1
-                for sg_group, date in list(pool_updates.items()):
-                    if pool_ports.get(sg_group):
-                        if min_date == -1 or date < min_date:
-                            min_date = date
-                            min_sg_group = sg_group
-                if min_date == -1:
-                    # pool is empty, no port to reuse
-                    raise exceptions.ResourceNotReady(pod)
-                port_id = pool_ports[min_sg_group].pop()
-            os_net.update_port(port_id, security_groups=list(security_groups))
-        if config.CONF.kubernetes.port_debug:
-            os_net.update_port(port_id, name=c_utils.get_port_name(pod))
-        # check if the pool needs to be populated
-        if (self._get_pool_size(pool_key) <
-                oslo_cfg.CONF.vif_pool.ports_pool_min):
-            eventlet.spawn(self._populate_pool, pool_key, pod, subnets,
-                           security_groups)
-        # Add protection from port_id not in existing_vifs
-        try:
-            port = self._existing_vifs[port_id]
-        except KeyError:
-            LOG.debug('Missing port on existing_vifs, this should not happen.'
-                      ' Retrying.')
-            raise exceptions.ResourceNotReady(pod)
-        return port
+        os_net.update_port(port_id, name=c_utils.get_port_name(pod))
 
     def _return_ports_to_pool(self):
         """Recycle ports to be reused by future pods.
@@ -967,9 +936,11 @@ class NestedVIFPool(BaseVIFPool):
                                     "reused, put back on the cleanable "
                                     "pool.", port_id)
                         continue
-                self._available_ports_pools.setdefault(
-                    pool_key, {}).setdefault(
-                        sg_current.get(port_id), []).append(port_id)
+                sg = sg_current.get(port_id)
+                self._available_ports_pools[pool_key].setdefault(
+                    sg, []).append(port_id)
+                # Move it to the end of ports to update the SG.
+                self._available_ports_pools[pool_key].move_to_end(sg)
             else:
                 trunk_id = self._get_trunk_id(pool_key)
                 try:
@@ -1074,10 +1045,9 @@ class NestedVIFPool(BaseVIFPool):
                         kuryr_subport, subnet, subport['segmentation_id'])
 
                     self._existing_vifs[kuryr_subport.id] = vif
-                    self._available_ports_pools.setdefault(
-                        pool_key, {}).setdefault(tuple(sorted(
-                            kuryr_subport.security_group_ids)),
-                            []).append(kuryr_subport.id)
+                    self._available_ports_pools[pool_key].setdefault(
+                        tuple(sorted(kuryr_subport.security_group_ids)),
+                        []).append(kuryr_subport.id)
 
                 elif action == 'free':
                     try:
@@ -1105,16 +1075,28 @@ class NestedVIFPool(BaseVIFPool):
         if not self._recovered_pools:
             LOG.debug("Kuryr-controller not yet ready to populate pools.")
             raise exceptions.ResourceNotReady(trunk_ip)
+
         pool_key = self._get_pool_key(trunk_ip, project_id, None, subnets)
-        pools = self._available_ports_pools.get(pool_key)
-        if not pools:
-            # NOTE(ltomasbo): If the amount of nodes is large the repopulation
-            # actions may take too long. Using half of the batch to prevent
-            # the problem
-            num_ports = int(max(oslo_cfg.CONF.vif_pool.ports_pool_batch/2,
-                            oslo_cfg.CONF.vif_pool.ports_pool_min))
-            self.force_populate_pool(trunk_ip, project_id, subnets,
-                                     security_groups, num_ports)
+        lock = self._get_populate_pool_lock(pool_key)
+
+        if lock.acquire(timeout=POPULATE_POOL_TIMEOUT):
+            try:
+                pools = self._available_ports_pools.get(pool_key)
+                if not pools:
+                    # NOTE(ltomasbo): If the amount of nodes is large the
+                    # repopulation actions may take too long. Using half of the
+                    # batch to prevent the problem
+                    num_ports = int(max(oslo_cfg.CONF.vif_pool
+                                        .ports_pool_batch/2,
+                                        oslo_cfg.CONF.vif_pool.ports_pool_min))
+                    self.force_populate_pool(trunk_ip, project_id, subnets,
+                                             security_groups, num_ports)
+            finally:
+                lock.release()
+        else:
+            LOG.debug("Kuryr-controller timed out waiting for it turn to "
+                      "populate pool, retrying.")
+            raise exceptions.ResourceNotReady(trunk_ip)
 
     def force_populate_pool(self, trunk_ip, project_id, subnets,
                             security_groups, num_ports=None):
@@ -1132,12 +1114,13 @@ class NestedVIFPool(BaseVIFPool):
             subnets=subnets,
             security_groups=security_groups,
             num_ports=num_ports,
-            trunk_ip=trunk_ip)
+            trunk_ip=trunk_ip,
+            semaphore=self._create_ports_semaphore)
 
         pool_key = self._get_pool_key(trunk_ip, project_id, None, subnets)
         for vif in vifs:
             self._existing_vifs[vif.id] = vif
-            self._available_ports_pools.setdefault(pool_key, {}).setdefault(
+            self._available_ports_pools[pool_key].setdefault(
                 tuple(sorted(security_groups)), []).append(vif.id)
 
     def free_pool(self, trunk_ips=None):
@@ -1154,7 +1137,10 @@ class NestedVIFPool(BaseVIFPool):
             LOG.debug("Kuryr-controller not yet ready to delete network "
                       "pools.")
             raise exceptions.ResourceNotReady(net_id)
-        os_net = clients.get_network_client()
+
+        epool = eventlet.GreenPool(constants.LEFTOVER_RM_POOL_SIZE)
+        ports_to_remove = []
+
         # NOTE(ltomasbo): Note the pods should already be deleted, but their
         # associated ports may not have been recycled yet, therefore not being
         # on the available_ports_pools dict. The next call forces it to be on
@@ -1168,10 +1154,14 @@ class NestedVIFPool(BaseVIFPool):
                         for p_id in sg_ports]
             try:
                 self._drv_vif._remove_subports(trunk_id, ports_id)
+            except os_exc.NotFoundException:
+                # We don't know which subport was already removed, but we'll
+                # attempt a manual detach on DELETE error, so just continue.
+                pass
             except (os_exc.SDKException, os_exc.HttpException):
                 LOG.exception('Error removing subports from trunk: %s',
                               trunk_id)
-                continue
+                raise exceptions.ResourceNotReady(net_id)
 
             for port_id in ports_id:
                 try:
@@ -1180,9 +1170,23 @@ class NestedVIFPool(BaseVIFPool):
                     del self._existing_vifs[port_id]
                 except KeyError:
                     LOG.debug('Port %s is not in the ports list.', port_id)
-                os_net.delete_port(port_id)
+                ports_to_remove.append(port_id)
 
-            self._available_ports_pools[pool_key] = {}
+            del self._available_ports_pools[pool_key]
+            with self._lock:
+                try:
+                    del self._populate_pool_lock[pool_key]
+                except KeyError:
+                    pass
+
+        # Parallelize Ports deletion. At this point the Ports
+        # should have been detatched from Trunk and if not operation
+        # will be retried
+        for result in epool.imap(c_utils.delete_neutron_port, ports_to_remove):
+            if result:
+                LOG.error('During Neutron port deletion an error occured: %s',
+                          result)
+                raise exceptions.ResourceNotReady(net_id)
 
 
 class MultiVIFPool(base.VIFPoolDriver):
@@ -1222,7 +1226,16 @@ class MultiVIFPool(base.VIFPoolDriver):
             self._vif_drvs[pod_driver] = drv_pool
 
     def request_vif(self, pod, project_id, subnets, security_groups):
-        pod_vif_type = self._get_pod_vif_type(pod)
+        pod_info = "%s/%s" % (pod['metadata']['namespace'],
+                              pod['metadata']['name'])
+        try:
+            pod_vif_type = self._get_pod_vif_type(pod)
+        except KeyError:
+            # NOTE(maysams): No nodeName set. Event should be skipped
+            LOG.warning("Pod %s has no .spec.nodeName set. This is unexpected "
+                        "as it's supposed to be scheduled. Ignoring event.",
+                        pod_info)
+            return None
         return self._vif_drvs[pod_vif_type].request_vif(
             pod, project_id, subnets, security_groups)
 
@@ -1230,9 +1243,9 @@ class MultiVIFPool(base.VIFPoolDriver):
         vif_drv_alias = self._get_vif_drv_alias(vif)
         self._vif_drvs[vif_drv_alias].release_vif(pod, vif, *argv)
 
-    def activate_vif(self, vif):
+    def activate_vif(self, vif, **kwargs):
         vif_drv_alias = self._get_vif_drv_alias(vif)
-        self._vif_drvs[vif_drv_alias].activate_vif(vif)
+        self._vif_drvs[vif_drv_alias].activate_vif(vif, **kwargs)
 
     def update_vif_sgs(self, pod, sgs):
         pod_vif_type = self._get_pod_vif_type(pod)
